import Data.List (sort)
import Data.Function (on)
import Control.Parallel.Strategies (parMap, rpar)
import Data.List.Split (splitOn)
import Data.List.Split (chunksOf)

type Position = (Int, Int)
type Maze = [[Int]]

data Node = Node
  { position :: Position
  , gCost :: Int
  , hCost :: Int
  , parent :: Maybe Node
  }

instance Eq Node where
  (==) = (==) `on` position

instance Ord Node where
  compare = compare `on` (\node -> gCost node + hCost node)

initialNode :: Maze -> Node
initialNode maze = Node {position = (0, 0), gCost = 0, hCost = heuristic maze (0, 0), parent = Nothing}

isGoal :: Maze -> Position -> Bool
isGoal maze pos = pos == (length maze - 1, length (head maze) - 1)

expandNode :: Maze -> Node -> [Node]
expandNode maze node =
  let (x, y) = position node
      neighbors = filter (\(dx, dy) -> isValidPosition maze (x + dx, y + dy)) [(1, 0), (-1, 0), (0, 1), (0, -1)]
  in map (\(dx, dy) -> Node {position = (x + dx, y + dy), gCost = 0, hCost = 0, parent = Just node}) neighbors

updateCosts :: Maze -> Node -> [Node] -> [Node]
updateCosts maze parent nodes =
  map
    (\node ->
       let g = gCost parent + 1
           h = heuristic maze (position node)
       in node {gCost = g, hCost = h, parent = Just parent})
    nodes

-- Heuristic function: Manhattan Distance
heuristic :: Maze -> Position -> Int
heuristic maze (x, y) = abs (x - goalX) + abs (y - goalY)
  where
    goalX = length maze - 1
    goalY = length (head maze) - 1

isValidPosition :: Maze -> Position -> Bool
isValidPosition maze (x, y) =
  x >= 0 && y >= 0 && x < length maze && y < length (head maze) && (maze !! x !! y) == 1

extractPath :: Node -> [Position]
extractPath node = go node []
  where
    go (Node pos _ _ (Just parent)) acc = go parent (pos : acc)
    go _ acc = (0, 0) : acc

-- Parallel processing of top 5 nodes in priority queue
astar5 :: Maze -> Maybe [Position]
astar5 maze = case go [initialNode maze] [] of
  Just node -> Just (extractPath node)
  Nothing   -> Nothing
  where
    go [] _ = Nothing
    go (current:rest) visited
      | isGoal maze (position current) = Just current
      | otherwise =
        let neighbors = filter (\n -> notElem (position n) visited) (expandNode maze current)
            newNodes = updateCosts maze current neighbors
            sortedNodes = sort (rest ++ newNodes)
            top5 = take 5 sortedNodes  -- Select the top 5 nodes
            expandedNodes = parMap rpar (\n -> go [n] (position current : visited)) top5
        in foldr (\result acc -> case result of Just node' -> Just node'; Nothing -> acc) Nothing expandedNodes

divideIntoArray :: String -> [[Int]]
divideIntoArray mazeString = chunksOf 100 (map read (splitOn "," mazeString))

printPath :: Maybe [Position] -> IO ()
printPath Nothing = putStrLn "No path found."
printPath (Just path) = putStrLn $ "Path: " ++ show path

---------------------------------------------------------------------------------------------------
main :: IO ()
main = do  
  content <- readFile "maze_examples.txt" 
  let mazeStrings = lines content 
      mazes = map divideIntoArray mazeStrings 
      solution = map astar5 mazes 

  mapM_ printPath solution  