import Data.List (sort)
import Data.Function (on)
import Data.List.Split (splitOn)
import Data.List.Split (chunksOf)
import qualified Test.HUnit as HUnit

type Position = (Int, Int)
type Maze = [[Int]]

data Node = Node
  { position :: Position
  , gCost :: Int
  , hCost :: Int
  , parent :: Maybe Node
  }

instance Eq Node where
  (==) = (==) `on` position

instance Ord Node where
  compare = compare `on` (\node -> gCost node + hCost node)

initialNode :: Maze -> Node
initialNode maze = Node {position = (0, 0), gCost = 0, hCost = heuristic maze (0, 0), parent = Nothing}

isGoal :: Maze -> Position -> Bool
isGoal maze pos = pos == (length maze - 1, length (head maze) - 1)

expandNode :: Maze -> Node -> [Node]
expandNode maze node =
  let (x, y) = position node
      neighbors = filter (\(dx, dy) -> isValidPosition maze (x + dx, y + dy)) [(1, 0), (-1, 0), (0, 1), (0, -1)]
  in map (\(dx, dy) -> Node {position = (x + dx, y + dy), gCost = 0, hCost = 0, parent = Just node}) neighbors

updateCosts :: Maze -> Node -> [Node] -> [Node]
updateCosts maze parent nodes =
  map
    (\node ->
       let g = gCost parent + 1
           h = heuristic maze (position node)
       in node {gCost = g, hCost = h, parent = Just parent})
    nodes

-- Heuristic function: Manhattan Distance
heuristic :: Maze -> Position -> Int
heuristic maze (x, y) = abs (x - goalX) + abs (y - goalY)
  where
    goalX = length maze - 1
    goalY = length (head maze) - 1

isValidPosition :: Maze -> Position -> Bool
isValidPosition maze (x, y) =
  x >= 0 && y >= 0 && x < length maze && y < length (head maze) && (maze !! x !! y) == 1

extractPath :: Node -> [Position]
extractPath node = go node []
  where
    go (Node pos _ _ (Just parent)) acc = go parent (pos : acc)
    go _ acc = (0, 0) : acc

-- Sequential A*
astar :: Maze -> Maybe [Position]
astar maze = case go [initialNode maze] [] of
  Just node -> Just (extractPath node)
  Nothing   -> Nothing
  where
    go [] _ = Nothing
    go (current:rest) visited
      | isGoal maze (position current) = Just current
      | otherwise =
        let neighbors = filter (\n -> notElem (position n) visited) (expandNode maze current)
            newNodes = updateCosts maze current neighbors
            sortedNodes = sort (rest ++ newNodes)
        in go sortedNodes (position current : visited)

---------------------- Sample mazes for testing ----------------------
-- Example 1: see if A* can find the SHORTEST path
sampleMaze1 :: Maze
sampleMaze1 =
  [ [1, 1, 1, 1, 1],
    [1, 0, 1, 0, 1],
    [1, 0, 1, 0, 1],
    [1, 0, 1, 0, 1],
    [1, 1, 1, 0, 1]
  ]
-- Solution 1
sampleSolution1 :: [Position]
sampleSolution1 = [(0,0),(0,1),(0,2),(0,3),(0,4),(1,4),(2,4),(3,4),(4,4)]
-- another path but not the shortest
-- (0,0),(1,0),(2,0),(3,0),(4,0),(4,1),(4,2),(3,2),(2,2),(1,2),(0,2),(0,3),(0,4),(1,4),(2,4),(3,4),(4,4)

-- Example 2
sampleMaze2 :: Maze
sampleMaze2 =
  [ [1, 0, 1, 1, 1],
    [1, 0, 1, 0, 1],
    [1, 0, 1, 0, 1],
    [1, 1, 1, 0, 1],
    [1, 0, 1, 0, 1]
  ]
-- Solution 2
sampleSolution2 :: [Position]
sampleSolution2 = [(0,0),(1,0),(2,0),(3,0),(3,1),(3,2),(2,2),(1,2),(0,2),(0,3),(0,4),(1,4),(2,4),(3,4),(4,4)]

-- Example 3: see if A* can find a path in a more complicated maze, and move in all 4 directions if possible
sampleMaze3 :: Maze
sampleMaze3 =
  [ [1, 1, 1, 1, 1, 1, 0, 1, 1, 0],
    [0, 0, 1, 1, 0, 1, 1, 0, 1, 0],
    [1, 1, 1, 0, 1, 0, 1, 1, 0, 1],
    [1, 1, 0, 1, 1, 1, 1, 1, 1, 1],
    [0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
    [1, 0, 1, 0, 1, 1, 1, 0, 1, 0],
    [0, 1, 1, 1, 0, 0, 1, 1, 0, 1],
    [1, 0, 0, 0, 1, 1, 1, 1, 1, 0],
    [0, 1, 1, 1, 1, 0, 0, 0, 0, 1],
    [1, 0, 0, 0, 1, 1, 1, 1, 1, 1] ]
-- Solution 3
sampleSolution3 :: [Position]
sampleSolution3 = [(0,0),(0,1),(0,2),(0,3),(0,4),(0,5),(1,5),(1,6),(2,6),(3,6),(3,5),(4,5),(5,5),(5,6),(6,6),(7,6),(7,5),(7,4),(8,4),(9,4),(9,5),(9,6),(9,7),(9,8),(9,9)]

-- Example 4: see the edge case where there are no paths in a maze
sampleMaze4 :: Maze
sampleMaze4 =
  [ [1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0],
    [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],
    [1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0],
    [1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0],
    [1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0],
    [1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0],
    [1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0],
    [1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1],
    [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1],
    [1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1]]
-- Solution 4 is Nothing

----------------------------------------------------------------------

-- Define test cases with an expected solution
testCase1 :: HUnit.Test
testCase1 = HUnit.TestCase $ HUnit.assertEqual
  "Test Case 1"
  (Just sampleSolution1)
  (astar sampleMaze1)

testCase2 :: HUnit.Test
testCase2 = HUnit.TestCase $ HUnit.assertEqual
  "Test Case 2"
  (Just sampleSolution2)
  (astar sampleMaze2)

testCase3 :: HUnit.Test
testCase3 = HUnit.TestCase $ HUnit.assertEqual
  "Test Case 3"
  (Just sampleSolution3)
  (astar sampleMaze3)

testCase4 :: HUnit.Test
testCase4 = HUnit.TestCase $ HUnit.assertEqual
  "Test Case 4"
  Nothing
  (astar sampleMaze4)



main :: IO ()
main = do
  -- Add your test cases here
  let testSuite = HUnit.TestList [testCase1, testCase2, testCase3, testCase4]

  -- Run the tests
  HUnit.runTestTT testSuite >>= print
